package eu.eudat.logic.managers;

import eu.eudat.types.MetricNames;
import eu.eudat.data.dao.criteria.*;
import eu.eudat.data.entities.Dataset;
import eu.eudat.data.entities.DatasetProfile;
import eu.eudat.logic.services.ApiContext;
import eu.eudat.logic.utilities.helpers.StreamDistinctBy;
import io.micrometer.prometheus.PrometheusMeterRegistry;
import io.prometheus.client.Gauge;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.env.Environment;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.transaction.Transactional;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Component
public class MetricsManager {
    private final static Logger logger = LoggerFactory.getLogger(MetricsManager.class);
    private final Map<String, Gauge> gauges;

    public void increaseValue(String name, int amount,  String label) {

        if(label != null) {
            gauges.get(name).labels(label).inc(amount);
        } else {
            gauges.get(name).inc(amount);
        }
    }
    
    public void decreaseValue(String name, int amount, String label) {
        if(label != null) {
            gauges.get(name).labels(label).dec(amount);
        } else {
            gauges.get(name).dec(amount);
        }
    }

    public Integer getValue(String name, String label) {
        if(label != null) {
            return Double.valueOf(gauges.get(name).labels(label).get()).intValue();
        } else {
            return Double.valueOf(gauges.get(name).get()).intValue();
        }
    }

    public void calculateValue(String name, int amount, String label) {
        Integer orig = getValue(name, label);
        int diff = orig - amount;
        if (diff != 0) {
            if (diff > 0) {
                decreaseValue(name, diff, label);
            } else {
                increaseValue(name, Math.abs(diff), label);
            }
        }
    }

    private final ApiContext apiContext;
    private final UserManager userManager;
    private final Environment environment;

    @Autowired
    public MetricsManager(ApiContext apiContext, UserManager userManager, Environment environment, PrometheusMeterRegistry registry) {
        this.apiContext = apiContext;
        this.userManager = userManager;
        this.environment = environment;
        registry.clear();
        this.gauges = Stream.of( new Object[][]{
                {MetricNames.DMP, Gauge.build().name("argos_managed_dmps").help("Number of managed DMPs").labelNames("status").register(registry.getPrometheusRegistry())},

                {MetricNames.FUNDERS, Gauge.build().name("argos_funders").help("Number of registered Funders").register(registry.getPrometheusRegistry())},
                {MetricNames.GRANTS, Gauge.build().name("argos_grants").help("Number of registered Grants").register(registry.getPrometheusRegistry())},
                {MetricNames.PROJECT, Gauge.build().name("argos_projects").help("Number of registered Projects").register(registry.getPrometheusRegistry())},
                {MetricNames.RESEARCHER, Gauge.build().name("argos_researchers").help("Number of Colaborators/Researchers").register(registry.getPrometheusRegistry())},

                {MetricNames.DATASET, Gauge.build().name("argos_managed_dataset_descriptions").help("Number of managed Dataset Descriptions").labelNames("status").register(registry.getPrometheusRegistry())},

                {MetricNames.DATASET_TEMPLATE, Gauge.build().name("argos_dataset_templates").help("Number of dataset Templates").labelNames("status").register(registry.getPrometheusRegistry())},

                {MetricNames.USERS, Gauge.build().name("argos_users").help("Number of users").labelNames("type").register(registry.getPrometheusRegistry())},

                {MetricNames.LANGUAGES, Gauge.build().name("argos_languages").help("Number of Languages").register(registry.getPrometheusRegistry())},

        }).collect(Collectors.toMap(data -> (String)data[0], data -> (Gauge) data[1]));

    }

    @PostConstruct
    @Transactional
    @Scheduled(initialDelay = 1000 * 60 * 60, fixedDelay = 1000 * 60 * 60)
    public void init() throws IOException {
        logger.info("Start calculating Metrics");
        calculateValue("argos_managed_dmps", (int) countAllDraftDMPs(), "draft");
        calculateValue("argos_managed_dmps", (int) countAllFinalizedDMPs(), "finalized");
        calculateValue("argos_managed_dmps", (int) countAllPublishedDMPs(), "published");
        calculateValue("argos_managed_dmps", (int) countAllDoiedDMPs(), "doied");

        calculateValue("argos_funders", (int) countAllFunders(), null);
        calculateValue("argos_grants", (int) countAllGrants(), null);
        calculateValue("argos_projects", (int) countAllProjects(), null);
        calculateValue("argos_researchers", (int) countAllResearchers(), null);

        calculateValue("argos_managed_dataset_descriptions", (int) countAllDraftDatasets(), "draft");
        calculateValue("argos_managed_dataset_descriptions", (int) countAllFinalizedDatasets(), "finalized");
        calculateValue("argos_managed_dataset_descriptions", (int) countAllPublicDatasets(), "published");
        calculateValue("argos_managed_dataset_descriptions", (int) countAllDatasetsWithDoi(), "doied");

        calculateValue("argos_dataset_templates", (int) countAllDraftTemplates(), "draft");
        calculateValue("argos_dataset_templates", (int) countAllFinalizedTemplates(), "active");
        calculateValue("argos_dataset_templates", (int) countAllUsedTemplates(), "used");

        calculateValue("argos_users", (int) userManager.countActiveUsers().intValue(), "loggedin");
        calculateValue("argos_users", (int) userManager.countAllUsers().intValue(), "total");

        long files = Files.list(Paths.get(this.environment.getProperty("userguide.path"))).count();
        calculateValue("argos_languages", (int) files, null);

        logger.info("Metrics calculation Completed");
    }

    private long countAllDraftDMPs() {
        DataManagementPlanCriteria criteria = new DataManagementPlanCriteria();
        criteria.setStatus(0);
        return apiContext.getOperationsContext().getDatabaseRepository().getDmpDao().getWithCriteria(criteria).count();
    }

    private long countAllFinalizedDMPs() {
        DataManagementPlanCriteria criteria = new DataManagementPlanCriteria();
        criteria.setStatus(1);
        return apiContext.getOperationsContext().getDatabaseRepository().getDmpDao().getWithCriteria(criteria).count();
    }

    private long countAllPublishedDMPs() {
        DataManagementPlanCriteria criteria = new DataManagementPlanCriteria();
        criteria.setIsPublic(true);
        criteria.setOnlyPublic(true);
        return apiContext.getOperationsContext().getDatabaseRepository().getDmpDao().getWithCriteria(criteria).count();
    }

    private long countAllDoiedDMPs() {
        DataManagementPlanCriteria criteria = new DataManagementPlanCriteria();
        criteria.setHasDoi(true);
        return apiContext.getOperationsContext().getDatabaseRepository().getDmpDao().getWithCriteria(criteria).count();
    }

    private long countAllResearchers() {
        ResearcherCriteria criteria = new ResearcherCriteria();
        return apiContext.getOperationsContext().getDatabaseRepository().getResearcherDao().getWithCriteria(criteria).count();
    }

    private long countAllProjects() {
        ProjectCriteria criteria = new ProjectCriteria();
        return apiContext.getOperationsContext().getDatabaseRepository().getProjectDao().getWithCritetia(criteria).count();
    }

    private long countAllFunders() {
        FunderCriteria criteria = new FunderCriteria();
        return apiContext.getOperationsContext().getDatabaseRepository().getFunderDao().getWithCritetia(criteria).count();
    }

    private long countAllGrants() {
        GrantCriteria criteria = new GrantCriteria();
        return apiContext.getOperationsContext().getDatabaseRepository().getGrantDao().getWithCriteria(criteria).count();
    }

    public long countAllDraftDatasets() {
        eu.eudat.data.dao.criteria.DatasetCriteria criteria = new eu.eudat.data.dao.criteria.DatasetCriteria();
        criteria.setStatus(0);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetDao().getWithCriteria(criteria).count();
    }

    public long countAllFinalizedDatasets() {
        eu.eudat.data.dao.criteria.DatasetCriteria criteria = new eu.eudat.data.dao.criteria.DatasetCriteria();
        criteria.setStatus(1);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetDao().getWithCriteria(criteria).count();
    }

    public long countAllPublicDatasets() {
        eu.eudat.data.dao.criteria.DatasetCriteria criteria = new eu.eudat.data.dao.criteria.DatasetCriteria();
        criteria.setIsPublic(true);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetDao().getWithCriteria(criteria).count();
    }

    public long countAllDatasetsWithDoi() {
        eu.eudat.data.dao.criteria.DatasetCriteria criteria = new eu.eudat.data.dao.criteria.DatasetCriteria();
        criteria.setHasDoi(true);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetDao().getWithCriteria(criteria).count();
    }

    public long countAllDraftTemplates() {
        DatasetProfileCriteria criteria = new DatasetProfileCriteria();
        criteria.setStatus(0);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetProfileDao().getWithCriteria(criteria).count();
    }

    public long countAllFinalizedTemplates() {
        DatasetProfileCriteria criteria = new DatasetProfileCriteria();
        criteria.setStatus(1);
        return apiContext.getOperationsContext().getDatabaseRepository().getDatasetProfileDao().getWithCriteria(criteria).count();
    }

    @Transactional
    public long countAllUsedTemplates() {
        DatasetProfileCriteria criteria = new DatasetProfileCriteria();
        criteria.setStatus(1);
        criteria.setAllVersions(false);
        List<DatasetProfile> datasetProfiles = apiContext.getOperationsContext().getDatabaseRepository().getDatasetProfileDao().getWithCriteria(criteria).toList();
        List<DatasetProfile> filteredProfiles = new ArrayList<>();
        for (DatasetProfile datasetProfile : datasetProfiles) {
            DatasetCriteria datasetCriteria = new DatasetCriteria();
            datasetCriteria.setDatasetTemplates(Collections.singletonList(datasetProfile.getId()));
            Long datasetCount = apiContext.getOperationsContext().getDatabaseRepository().getDatasetDao().getWithCriteria(datasetCriteria).count();
            if (datasetCount > 0) {
                filteredProfiles.add(datasetProfile);
            }
        }
        return filteredProfiles.stream().filter(StreamDistinctBy.distinctByKey(DatasetProfile::getId)).count();
    }

}
